import json
import os
import time
from typing import Iterable
from PIL import Image
from PIL.PngImagePlugin import PngInfo
from pixaris.experiment_handlers.base import ExperimentHandler
import pandas as pd


class LocalExperimentHandler(ExperimentHandler):
    def __init__(self, local_results_folder: str = "local_results"):
        """
        Initialize the LocalExperimentHandler.
        Args:
            local_results_folder (str, optional): The root folder where the experiment subfolder is located. Defaults to 'local_results'.
        """
        self.local_results_folder = local_results_folder

    def store_results(
        self,
        project: str,
        dataset: str,
        experiment_run_name: str,
        image_name_pairs: Iterable[tuple[Image.Image, str]],
        metric_values: dict[str, float],
        args: dict[str, any] = {},
        dataset_tracking_file_name: str = "experiment_tracking.jsonl",
    ):
        """
        Save a collection of images locally under a specified experiment name.

        :param project: The name of the project. This will be used to create a subfolder where images will be saved.
        :type project: str
        :param dataset: The name of the evaluation set.
        :type dataset: str
        :param experiment_run_name: The name of the experiment. This will be used to create a subfolder where images will be saved.
        :type experiment_run_name: str
        :param image_name_pairs: An iterable collection of tuples, where each tuple contains a PIL Image object and its corresponding name.
        :type image_name_pairs: Iterable[tuple[Image.Image, str]]
        :param metric_values: The metrics of the experiment to be saved as a JSON file.
        :type metric_values: dict[str, float]
        :param args: The arguments of the experiment to be saved as a JSON file. If any argument is a PIL Image, it will be saved as an image file.
        :type args: dict[str, any]
        :param local_results_folder: The root folder where the experiment subfolder will be created. Defaults to 'local_results'.
        :type local_results_folder: str, optional
        :param global_tracking_file: The name of the global tracking file. Defaults to 'all_experiment_results.jsonl'.
        :type global_tracking_file: str, optional
        """
        timestamp = time.strftime("%Y%m%d-%H%M%S")
        save_dir = os.path.join(
            self.local_results_folder,
            project,
            dataset,
            timestamp + "_" + experiment_run_name,
        )

        os.makedirs(save_dir, exist_ok=True)

        # Save each image in the collection
        if image_name_pairs:
            os.makedirs(os.path.join(save_dir, "images"), exist_ok=True)
        for image, name in image_name_pairs:
            image.save(
                os.path.join(save_dir, "images", name.split(".")[0] + ".png"),
                "PNG",
                # if you switch to JPEG, use quality=95 as input! Otherwise, expect square artifacts
            )

        args_with_files_as_paths = {}
        for key, value in args.items():
            if isinstance(value, Image.Image):
                metadata = PngInfo()
                for metadata_key, metadata_value in value.info.items():
                    metadata.add_text(metadata_key, str(metadata_value))
                image_path = os.path.join(save_dir, key + ".png")
                value.save(image_path, pnginfo=metadata)
                args_with_files_as_paths[key] = image_path
            elif isinstance(value, dict):
                # If the value is a dictionary, save it as a JSON file
                json_path = os.path.join(save_dir, key + ".json")
                with open(json_path, "w") as f:
                    json.dump(value, f)
                args_with_files_as_paths[key] = json_path
            else:
                args_with_files_as_paths[key] = value

        # build experiment tracking info json
        tracking_info = {
            "timestamp": timestamp,
            "experiment_run_name": experiment_run_name,
        }
        tracking_info.update(args_with_files_as_paths)
        tracking_info.update(metric_values)

        # Save the results as JSON files in the experiment subfolder
        with open(os.path.join(save_dir, "args.json"), "w") as f:
            json.dump(tracking_info, f)

        # Append the results to the global tracking file
        with open(
            os.path.join(
                self.local_results_folder, project, dataset, dataset_tracking_file_name
            ),
            "a",
        ) as f:
            f.write(json.dumps(tracking_info) + "\n")

    def load_projects_and_datasets(
        self,
    ):
        """
        Load the projects and datasets from the local results folder.
        Returns:
            dict: A dictionary containing the projects and datasets. {"project": ["dataset1", "dataset2"]}
        """
        projects = os.listdir(self.local_results_folder)
        projects.sort()
        project_dict = {}
        for project in projects:
            project_path = os.path.join(self.local_results_folder, project)
            if os.path.isdir(project_path):
                project_dict[project] = os.listdir(project_path)
        return project_dict

    def load_experiment_results_for_dataset(
        self,
        project: str,
        dataset: str,
    ) -> pd.DataFrame:
        """
        Load the results of an experiment.
        Args:
            project (str): The name of the project.
            dataset (str): The name of the evaluation set.
        Returns:
            pd.DataFrame: The results of the experiment as a DataFrame.
        """
        results_file = os.path.join(
            self.local_results_folder,
            project,
            dataset,
            "experiment_results.jsonl",
        )

        if os.path.exists(results_file) and os.stat(results_file).st_size > 0:
            try:
                return pd.read_json(results_file, lines=True)
            except ValueError:
                pass
        return pd.DataFrame()



from pixaris.feedback_handlers.base import FeedbackHandler
from google.cloud import bigquery, storage
import gradio as gr
from datetime import datetime
import os
from PIL import Image


class BigqueryFeedbackHandler(FeedbackHandler):
    def __init__(
        self,
        gcp_project_id: str,
        gcp_bq_feedback_table: str,
        gcp_feedback_bucket: str,
        local_feedback_directory: str = "local_results",
    ):
        self.gcp_project_id = gcp_project_id
        self.gcp_bq_feedback_table = gcp_bq_feedback_table
        self.gcp_feedback_bucket = gcp_feedback_bucket
        os.makedirs(local_feedback_directory, exist_ok=True)
        self.local_feedback_directory = local_feedback_directory
        self.feedback_df = None
        self.feedback_iteration_choices = None
        self.projects = None

    def write_single_feedback(self, feedback: dict) -> None:
        """
        Writes feedback for one image to BigQuery table.

        :param feedback: dict with feedback information. Dict is expected to have the following keys:
            - project: name of the project
            - feedback_iteration: name of the iteration
            - dataset: name of the evaluation set (optional)
            - image_name: name of the image
            - experiment_name: name of the experiment (optional)
            - feedback_indicator: string with feedback value (either "Like", "Dislike", or "Neither")
            - comment: string with feedback comment (optional)
        :type feedback: dict
        """
        # assert non-nullable values are present
        assert all(
            key in feedback.keys()
            for key in [
                "project",
                "feedback_iteration",
                "image_name",
                "feedback_indicator",
            ]
        ), (
            "Missing required feedback keys. Must have 'project', 'feedback_iteration', 'image_name', 'feedback_indicator'"
        )

        # setup row to insert to table
        row_to_insert = {
            "project": feedback["project"],
            "feedback_iteration": feedback["feedback_iteration"],
            "dataset": feedback.get("dataset", ""),
            "image_name": feedback["image_name"],
            "experiment_name": feedback.get("experiment_name", ""),
            "date": datetime.now().isoformat(),
            "comment": feedback.get("comment", ""),
        }

        # determine what to write to feedback columns
        feedback_indicator = feedback["feedback_indicator"]

        if feedback_indicator == "Like":
            row_to_insert["likes"] = 1
            row_to_insert["dislikes"] = 0
        elif feedback_indicator == "Dislike":
            row_to_insert["likes"] = 0
            row_to_insert["dislikes"] = 1
        elif (
            feedback_indicator == "Neither"
        ):  # Is used when uploading images (no feedback given)
            row_to_insert["likes"] = 0
            row_to_insert["dislikes"] = 0
        else:
            raise ValueError(
                "Invalid feedback indicator. Must be 'Like', 'Dislike', or 'Neither'"
            )

        client = bigquery.Client(project=self.gcp_project_id)
        errors = client.insert_rows_json(self.gcp_bq_feedback_table, [row_to_insert])

        # Check for errors and display warnings to UI
        if errors == []:  # todo: move displaying this to frontend!
            gr.Info("Feedback sent successfully", duration=1)
        else:
            gr.Warning(
                f"Errors occurred while inserting row: {errors[0]['errors']}",
                duration=10,
            )

    def load_projects_list(self) -> list[str]:
        """
        Retrieves list of projects from BigQuery table.

        Returns:
            List of project names.
        """
        print("Querying BigQuery for list of projects...")
        client = bigquery.Client(project=self.gcp_project_id)

        query = f"""
            SELECT
                DISTINCT `project`
            FROM
                {self.gcp_bq_feedback_table};
        """
        rows = client.query_and_wait(query)
        projects = rows.to_dataframe()["project"].tolist()
        projects.sort()
        self.projects = projects

        print(f"Done. Found projects: {projects}")
        return projects

    def load_all_feedback_iterations_for_project(self, project: str) -> None:
        """
        Retrieves feedback data for a project from BigQuery table. Adds paths for location of images in
        GCP bucket and local directory to the dataframe. Saves the resulting df to self.feedback_df.
        Saves the list of feedback iterations to self.feedback_iteration_choices.

        Args:
            project: str Name of the project

        Returns:
            None
        """
        print(f"Querying BigQuery for feedback data for project {project}...")
        client = bigquery.Client(project=self.gcp_project_id)

        query = f"""
            SELECT
                `project`,
                feedback_iteration,
                image_name,
                SUM(likes) AS likes_count,
                SUM(dislikes) AS dislikes_count,
                STRING_AGG(IF(likes > 0 AND comment <> "upload", comment, NULL), ', ') AS comments_liked,
                STRING_AGG(IF(dislikes > 0 AND comment <> "upload", comment, NULL), ', ') AS comments_disliked
            FROM
                {self.gcp_bq_feedback_table}

            WHERE
                `project` = "{project}"
            GROUP BY
                `project`,
                feedback_iteration,
                image_name;
        """
        rows = client.query_and_wait(query)
        df = rows.to_dataframe()

        # add paths for images to df (local and GCS bucket)
        df["image_path_bucket"] = (
            df["project"] + "/" + df["feedback_iteration"] + "/" + df["image_name"]
        )
        df["image_path_local"] = (
            f"{self.local_feedback_directory}/"
            + df["project"]
            + "/feedback_iterations/"
            + df["feedback_iteration"]
            + "/"
            + df["image_name"]
        )

        # determine feedback iterations to choose from in this project
        choices = df["feedback_iteration"].unique().tolist()
        choices.sort()

        self.feedback_iteration_choices = choices
        self.feedback_df = df

        print(f"Done. Found feedback iterations: {choices}")

    def load_images_for_feedback_iteration(
        self,
        feedback_iteration: str,
    ) -> list[str]:
        """
        Downloads images for a feedback iteration from GCP bucket to local directory.
        Returns list of local image paths that belong to the feedback iteration.

        Args:
            feedback_iteration: str Name of the feedback iteration

        Returns:
            List of local image paths.
        """
        print(f"Downloading images for feedback iteration {feedback_iteration}...")

        # get relevant data for this feedback iteration
        iteration_df = self.feedback_df.loc[
            self.feedback_df["feedback_iteration"] == feedback_iteration
        ].copy()

        # download images
        for row in iteration_df.iterrows():
            image_path_bucket = row[1]["image_path_bucket"]
            image_path_local = row[1]["image_path_local"]
            if not os.path.exists(image_path_local):
                gr.Info(f"Downloading image '{image_path_bucket}'...", duration=1)
                os.makedirs(os.path.dirname(image_path_local), exist_ok=True)
                self._download_image(image_path_bucket, image_path_local)

        print("Done.")
        return iteration_df["image_path_local"].tolist()

    def _download_image(
        self,
        image_path_bucket: str,
        image_path_local: str,
    ) -> None:
        storage_client = storage.Client(project=self.gcp_project_id)
        bucket = storage_client.bucket(self.gcp_feedback_bucket)

        # download image if possible, otherwise fill with white placeholder image
        try:
            blob = bucket.blob(image_path_bucket)
            blob.download_to_filename(image_path_local)
        except Exception as e:
            print(f"Error downloading image '{image_path_bucket}': {e}")
            print("Filling with placeholder image")
            Image.new(mode="RGB", size=(256, 256), color="white").save(image_path_local)



import gradio as gr
import os
from pixaris.frontend.tab_feedback import render_feedback_tab
from pixaris.frontend.tab_experiment_tracking import render_experiment_tracking_tab

from pixaris.feedback_handlers.base import FeedbackHandler
from pixaris.experiment_handlers.base import ExperimentHandler

from pixaris.feedback_handlers.local import LocalFeedbackHandler
from pixaris.experiment_handlers.local import LocalExperimentHandler

import yaml


def launch_ui(
    feedback_handler: FeedbackHandler,
    experiment_handler: ExperimentHandler,
    server_name="localhost",
):
    """
    Launches the Gradio UI for Pixaris.
    Args:
        feedback_handler: The feedback handler to use.
        experiment_handler: The experiment handler to use.
        server_name: The name of the server to launch the UI on. Set "localhost" for local testing and "0.0.0.0" for app engine deployment.
    """
    with gr.Blocks(
        title="Pixaris",
        theme=gr.themes.Default(
            spacing_size=gr.themes.sizes.spacing_sm, radius_size="none"
        ),
        analytics_enabled=False,
    ) as demo:
        results_directory = "local_results/"

        with gr.Tab("Experiment Tracking"):
            render_experiment_tracking_tab(
                experiment_handler=experiment_handler,
                results_directory=results_directory,
            )

        with gr.Tab("Feedback"):
            render_feedback_tab(
                feedback_handler=feedback_handler,
            )
    demo.launch(
        server_name=server_name, server_port=8080, allowed_paths=[os.path.abspath("./")]
    )


if __name__ == "__main__":
    config = yaml.safe_load(open("pixaris/config.yaml"))
    feedback_handler = LocalFeedbackHandler()

    # feedback_handler = BigqueryFeedbackHandler(
    #     gcp_project_id=config["gcp_project_id"],
    #     gcp_bq_feedback_table=config["gcp_bq_feedback_table"],
    #     gcp_feedback_bucket=config["gcp_feedback_bucket"],
    # )
    experiment_handler = LocalExperimentHandler()  # TODO use BigqueryExperimentTracker

    launch_ui(
        feedback_handler=feedback_handler,
        experiment_handler=experiment_handler,
        server_name="localhost",
    )



# from pixaris.experiment_handlers.utils import upload_workflow_file_to_bucket
# from typing import Iterable
# from PIL import Image
# from PIL.PngImagePlugin import PngInfo
# from google.api_core.exceptions import AlreadyExists
# from google.cloud import aiplatform
# import tensorflow as tf
# import os
# import shutil
# from datetime import datetime
# import numpy as np
# import json
# import re

# from pixaris.experiment_handlers.base import ExperimentHandler


# class GCPTensorboardHandler(ExperimentHandler):
#     def __init__(
#         self,
#         gcp_project_id: str,
#         location: str,
#         bucket_name: str = None,
#     ):
#         """
#         Initializes the GCPTensorboardHandler.

#         :param gcp_project_id: The Google Cloud project ID.
#         :type gcp_project_id: str
#         :param location: The Google Cloud location.
#         :type location: str
#         :param bucket_name: The name of the Google Cloud Storage bucket to store the workflow image. If None, the workflow image will not be stored in a bucket. Defaults to None.
#         :type bucket_name: str, optional
#         """
#         self.gcp_project_id = gcp_project_id
#         self.location = location
#         self.bucket_name = bucket_name

#     def _validate_experiment_run_name(
#         self,
#         experiment_run_name: str,
#     ) -> None:
#         """
#         assert experiment_run_name adheres to tensorboard rules
#         """
#         assert re.match(r"[a-z0-9][a-z0-9-]{0,127}", experiment_run_name), (
#             "experiment_run_name must adhere to regex [a-z0-9][a-z0-9-]{0,127} - so only lowercase letters, numbers and '-' are allowed, max length 128"
#         )

#     def _validate_args(self, args: dict[str, any]):
#         # check if all keys are strings
#         assert all(isinstance(key, str) for key in args.keys()), (
#             "All keys must be strings."
#         )

#         # check if "pillow_images" is a list of dictionaries containing the correct keys
#         if "pillow_images" in args:
#             pillow_images = args["pillow_images"]
#             assert isinstance(pillow_images, list), "pillow_images must be a list."
#             assert all(isinstance(item, dict) for item in pillow_images), (
#                 "Each item in the list must be a dictionary."
#             )
#             assert all(
#                 all(key in item for key in ["node_name", "pillow_image"])
#                 for item in pillow_images
#             ), "Each dictionary must contain the keys 'node_name' and 'pillow_image'."

#     def _save_args_entry(self, args: (dict[str, any])):
#         """
#         Saves all args to TensorBoard.
#             if value is a Pillow image, save as image
#             if value is a dictionary, save as text
#             if value is a number, save as scalar
#             if key is "pillow_images", save the images under their node names. Validity checked beforehand
#             else save value as json dump

#         :param args: A dictionary of arguments to be saved to TensorBoard.
#         :type args: dict[str, any]
#         """
#         # save all args depending on their type
#         for key, value in args.items():
#             if isinstance(value, Image.Image):
#                 tf.summary.image(
#                     key,
#                     [np.asarray(value) / 255],
#                     step=0,
#                 )

#             elif isinstance(value, dict):
#                 tf.summary.text(key, json.dumps(value), step=0)

#             elif isinstance(value, (int, float)):
#                 tf.summary.scalar(key, value, step=0)

#             elif key == "pillow_images":
#                 for pillow_image_dict in value:
#                     tf.summary.image(
#                         pillow_image_dict["node_name"],
#                         [np.asarray(pillow_image_dict["pillow_image"]) / 255],
#                         step=0,
#                     )

#             else:
#                 tf.summary.text(key, json.dumps(value), step=0)

#     def _save_workflow_image_to_bucket(
#         self, args: dict[str, any], dataset: str, experiment_run_name: str
#     ):
#         """Saves the workflow image to a bucket, keeping the metadata."""

#         workflow_pillow_image = args["workflow_pillow_image"]

#         workflow_image_path = os.path.abspath(
#             f"temp/logs/{dataset}/{experiment_run_name}/workflow_image.png"
#         )
#         metadata = PngInfo()
#         for key, value in workflow_pillow_image.info.items():
#             metadata.add_text(key, str(value))
#         workflow_pillow_image.save(workflow_image_path, pnginfo=metadata)
#         link_to_workflow_in_bucket = upload_workflow_file_to_bucket(
#             gcp_project_id=self.gcp_project_id,
#             bucket_name=self.bucket_name,
#             dataset=dataset,
#             experiment_run_name=experiment_run_name,
#             local_file_path=workflow_image_path,
#         )
#         return link_to_workflow_in_bucket

#     def store_results(
#         self,
#         dataset: str,
#         experiment_run_name: str,
#         image_name_pairs: Iterable[tuple[Image.Image, str]],
#         metric_values: dict[str, float],
#         args: dict[str, any] = {},
#         project: str = "",
#     ):
#         """
#         Stores the results of an evaluation run to TensorBoard.

#         :param dataset: The name of the evaluation set.
#         :type dataset: str
#         :param experiment_run_name: The name of the run.
#         :type experiment_run_name: str
#         :param image_name_pairs: A collection of image and name pairs to log.
#         :type image_name_pairs: Iterable[tuple[Image.Image, str]]
#         :param metric_values: A dictionary of metric names and their corresponding values.
#         :type metric_values: dict[str, float]
#         :param args: Arguments given to the ImageGenerator that generated the images. Defaults to an empty dictionary.
#         :type args: dict[str, any], optional
#         :raises AssertionError: If any value in the metrics dictionary is not a number.
#         """
#         self._validate_args(args)

#         aiplatform.init(
#             experiment=dataset.replace("_", "-"),
#             project=self.gcp_project_id,
#             location=self.location,
#             experiment_tensorboard=True,
#         )

#         # remove old logs
#         shutil.rmtree(os.path.abspath(f"temp/logs/{dataset}"), ignore_errors=True)

#         # add date to experiment_run_name in format YYMMDD-experiment_run_name
#         experiment_run_name = (
#             f"{datetime.now().strftime('%y%m%d')}-{experiment_run_name}"
#         )

#         # handle run name if it already exists in dataset
#         try:
#             aiplatform_run = aiplatform.start_run(experiment_run_name)
#         except Exception as e:
#             if type(e) is AlreadyExists:
#                 print(
#                     f"Run name {experiment_run_name} already exists in experiment {dataset}. Adding random number."
#                 )
#                 experiment_run_name = experiment_run_name + str(np.random.randint(99))
#                 print(f"New run name: {experiment_run_name}. Continuing.")
#                 aiplatform_run = aiplatform.start_run(experiment_run_name)
#             else:
#                 raise e

#         with aiplatform_run:
#             with tf.summary.create_file_writer(
#                 os.path.abspath(f"temp/logs/{dataset}/{experiment_run_name}"),
#             ).as_default():
#                 # save generated images
#                 print("Logging generated images")
#                 for image, name in image_name_pairs:
#                     tf.summary.image(
#                         name,
#                         [np.asarray(image) / 255],
#                         step=0,
#                     )

#                 # save metrics
#                 print("Logging metrics")
#                 for metric, value in metric_values.items():
#                     assert isinstance(value, (int, float)), (
#                         f"Value for metric {metric} is not a number."
#                     )
#                     tf.summary.scalar(metric, value, step=0)

#                 # save all args depending on their type
#                 print("Logging args")
#                 self._save_args_entry(args)

#                 # save workflow image with metadata to bucket
#                 if "workflow_pillow_image" in args and self.bucket_name:
#                     print("Saving workflow image to bucket")
#                     # Save the workflow image locally before uploading
#                     link_to_workflow_in_bucket = self._save_workflow_image_to_bucket(
#                         args=args,
#                         dataset=dataset,
#                         experiment_run_name=experiment_run_name,
#                     )
#                     tf.summary.text(
#                         "link_to_workflow_image", link_to_workflow_in_bucket, step=0
#                     )

#             aiplatform.upload_tb_log(
#                 tensorboard_experiment_name=dataset.replace("_", "-"),
#                 experiment_display_name=experiment_run_name,
#                 logdir=os.path.abspath(f"temp/logs/{dataset}"),
#             )



from concurrent.futures import ThreadPoolExecutor
import json
import re
import io
from PIL.Image import Image
import vertexai
from vertexai.generative_models import GenerativeModel, Image as GenAIImage, Part
import yaml
from pixaris.metrics.base import BaseMetric
from pixaris.metrics.utils import dict_mean


class LLMMetric(BaseMetric):
    def __init__(self, object_images: list[Image], style_images: list[Image]):
        super().__init__()
        self.object_images = object_images
        self.style_images = style_images

    def _PIL_image_to_vertex_image(self, image: Image) -> GenAIImage:
        """
        Converts a PIL image to a vertex image.
        Args:
            image (PIL.Image): The input PIL image.
        Returns:
            GenAIImage: The converted vertex image.
        """
        img_byte_arr = io.BytesIO()
        image.save(img_byte_arr, format="PNG")
        return GenAIImage.from_bytes(img_byte_arr.getvalue())

    def _llm_prompt(
        self,
        evaulation_image: Image,
        object_image: Image,
        style_image: Image,
    ):
        """
        Generates a prompt for rating images on various metrics and returns a JSON object with the ratings.
        Args:
            input_image_vertex (Part): The input image.
            output_image_vertex (Part): The output image.
            style_image_vertex (Part): The style image.
        Returns:
            str: A prompt that provokes a response  with a JSON with the following keys:
                - llm_reality: How real does the first picture look where 0 is not real at all and 1 is photorealistic?
                - llm_similarity: How similar is the main object in the first picture to the template in the second picture where 0 is not similar at all and 1 is identical?
                - llm_errors: How many errors can you find in the first picture?
                - llm_style: How well does the style of the first picture match the style of the third picture where 0 is not at all and 1 is identical?
        """
        json_prompt = "rate the following evluation image on the following metrics. return only a json file without newlines with the following keys:"
        reality_prompt = "llm_reality: How real does the evaluation image look where 0 is not real at all and 1 is photorealistic?"
        similarity_prompt = "llm_similarity: How similar is the main object in the evaluation image to the template in the following picture where 0 is not similar at all and 1 is identical?"
        error_prompt = (
            "llm_errors: How many errors can you find in the evaulation image?"
        )
        style_prompt = "llm_style: How well does the style of the evaluation image match the style of the following style image where 0 is not at all and 1 is identical?"

        return [
            json_prompt,
            Part.from_image(self._PIL_image_to_vertex_image(evaulation_image)),
            reality_prompt,
            similarity_prompt,
            Part.from_image(self._PIL_image_to_vertex_image(object_image)),
            error_prompt,
            style_prompt,
            Part.from_image(self._PIL_image_to_vertex_image(style_image)),
        ]

    def _postprocess_response(self, response_text: str) -> str:
        """
        If there is some sort of JSON-like structure in the response text, extract it and return it.
        Args:
        response_text (str): The response text from the model.
        Returns:
        str: The extracted JSON-like structure if found, otherwise the original response text.
        """
        pattern = r"\{.*\}"
        match = re.search(pattern, response_text)
        if match:
            extracted_string = match.group(0)
            return extracted_string
        else:
            raise ValueError("No JSON-like structure found in the llm response text.")

    def _response_to_dict(self, response_text: str) -> dict:
        """
        Converts the response text to a dictionary.
        Args:
            response_text (str): The response text from the model.
        Returns:
            dict: The response text as a dictionary.
        """
        parsed_text = self._postprocess_response(response_text)
        response_dict = json.loads(parsed_text)
        response_dict = {key: float(value) for key, value in response_dict.items()}

        if not all(
            metrics in response_dict
            for metrics in ["llm_reality", "llm_similarity", "llm_errors", "llm_style"]
        ):
            raise ValueError("Response dictionary does not contain all required keys.")
        return response_dict

    def _call_gemini(self, prompt) -> str:
        """
        Sends the prompt to Google API
        Args:
            prompt (str): The prompt for the LLM metrics. Generated by llm_prompt().
        Returns:
            string: The LLM response.
        """
        with open("pixaris/config.yaml", "r") as f:
            config = yaml.safe_load(f)

        vertexai.init(project=config["gcp_project_id"], location=config["gcp_location"])

        model = GenerativeModel("gemini-1.5-flash-preview-0514")

        responses = model.generate_content(prompt, stream=False)

        return responses.text

    def _successful_evaluation(self, prompt, max_tries: int = 3) -> dict:
        """
        Perform an evaluation by calling the `call_gemini` function with the given parameters.
        Assures that gemini returns correct json code by calling it up to max_tries times if it fails.
        Args:
            prompt: The prompt for the
            max_tries (optional): The maximum number of tries to perform the  Defaults to 3.
        Returns:
            A dictionary containing the response from the `call_gemini` function.
        Raises:
            ValueError: If an error occurs during the
        """
        for i in range(max_tries):
            try:
                return self._response_to_dict(self._call_gemini(prompt))
            except ValueError:
                pass

    def _combined_score(self, response: dict) -> float:
        """
        Calculates the combined score from the response dictionary.
        Args:
            response (dict): The response dictionary.
        Returns:
            float: The combined score. Score at the moment is just the mean of all the scores.
        """
        return sum(response.values()) / len(response.values())

    def _llm_scores_per_image(
        self,
        evaluation_image: Image,
        object_image: Image,
        style_image: Image,
        sample_size: int = 3,
    ) -> dict:
        """
        Calculates the LLM score for the generated image.
        Args:
            generated_image (Image): The generated image.
            template_image (Image): The template image.
            style_image (Image): The style image.
            sample_size (optional): The number of llm calls to evaluate. Defaults to 3
        Returns:
            dict: A dictionary containing the LLM scores.
        """
        scores = [
            self._successful_evaluation(
                self._llm_prompt(evaluation_image, object_image, style_image)
            )
            for _ in range(sample_size)
        ]

        # Calculate the average score for each metric
        average_scores_per_metric = dict_mean(scores)

        # Invert the error count to get a score
        average_scores_per_metric["llm_errors"] = 1 / (
            1 + average_scores_per_metric["llm_errors"]
        )

        # Calculate the llm_average score
        average_scores_per_metric["llm_average"] = self._combined_score(
            average_scores_per_metric
        )

        return average_scores_per_metric

    def calculate(self, x: list[Image]) -> dict:
        """
        Calculate the LLM metrics for a list of generated images.
        Args:
            x (list[Image]): A list of evaluation images.
        Returns:
            dict: A dictionary containing the mean LLM scores for each prompt.
        """
        if len(x) != len(self.object_images):
            raise ValueError(
                "There should be as many object images as generated images in the llm metric"
            )
        if len(x) != len(self.style_images):
            raise ValueError(
                "There should be as many style images as generated images in the llm metric"
            )

        with ThreadPoolExecutor(len(x)) as executor:
            llm_metrics = dict_mean(
                list(
                    executor.map(
                        self._llm_scores_per_image,
                        x,
                        self.object_images,
                        self.style_images,
                    )
                )
            )
            return llm_metrics



import gradio as gr
import os
from pixaris.experiment_handlers.base import ExperimentHandler
import pandas as pd


def render_experiment_tracking_tab(
    experiment_handler: ExperimentHandler,
    results_directory: str,
):
    dataset_experiment_tracking_results = gr.State(pd.DataFrame())
    with gr.Sidebar(open=True):
        gr.Markdown("Experiments")
        with gr.Row(scale=8):
            PROJECTS_DICT = experiment_handler.load_projects_and_datasets()
            PROJECTS = [""] + list(PROJECTS_DICT.keys())
            project_name = gr.Dropdown(
                choices=PROJECTS,
                value="",
                label="Project",
                filterable=True,
            )

            # initialise hidden feedback iterations and button
            dataset = gr.Dropdown(
                visible=False,
                value="",
                choices=[""],
            )

            def update_dataset_choices(project_name, dataset):
                """Update choices of feedback iterations for selected project and display reload button."""
                dataset_choices = PROJECTS_DICT[project_name]

                dataset = gr.Dropdown(
                    label="Dataset",
                    choices=dataset_choices,
                    filterable=True,
                    multiselect=False,
                    interactive=True,
                    visible=True,
                )
                return dataset

            project_name.change(
                fn=update_dataset_choices,
                inputs=[
                    project_name,
                    dataset,
                ],
                outputs=[dataset],
            )

            experiments = gr.Dropdown(
                visible=False,
                value="",
                choices=[""],
            )

            def update_experiments_choices(
                project_name, dataset, experiments, dataset_experiment_tracking_results
            ):
                """Update choices of feedback iterations for selected project and display reload button."""
                dataset_experiment_tracking_results = (
                    experiment_handler.load_experiment_results_for_dataset(
                        project=project_name,
                        dataset=dataset,
                    )
                )
                experiment_choices = list(
                    dataset_experiment_tracking_results["experiment_run_name"]
                    .dropna()
                    .unique()
                )
                experiment_choices.sort()
                experiments = gr.Dropdown(
                    choices=experiment_choices,
                    value=experiment_choices[-1],
                    label="Experiments",
                    filterable=True,
                    multiselect=True,
                    max_choices=100,
                    visible=True,
                )
                return experiments, dataset_experiment_tracking_results

            dataset.change(
                fn=update_experiments_choices,
                inputs=[
                    project_name,
                    dataset,
                    experiments,
                    dataset_experiment_tracking_results,
                ],
                outputs=[experiments, dataset_experiment_tracking_results],
            )

        with gr.Row(scale=1):
            columns = gr.Slider(
                minimum=1,
                maximum=20,
                value=8,
                label="Number of images per row",
                step=1,
            )
        with gr.Row(scale=1):
            gallery_height = gr.Slider(
                minimum=100,
                maximum=1000,
                value=360,
                label="Gallery height",
                step=10,
            )

    with gr.Tab("Images"):

        @gr.render(inputs=[project_name, dataset, experiments, columns, gallery_height])
        def show_gallery(project_name, dataset, experiments, columns, gallery_height):
            """Renders one gallery per experiment. Render decorator enables listening to experiments checkbox group."""
            if not experiments:
                gr.Markdown("No experiment chosen.")
            else:
                for experiment_name in experiments:
                    experiment_images = os.listdir(
                        os.path.join(
                            results_directory,
                            project_name,
                            dataset,
                            experiment_name,
                            "images",
                        )
                    )
                    gr.Gallery(
                        value=experiment_images,
                        label=experiment_name,
                        columns=columns,
                        rows=1,
                        show_download_button=True,
                        show_fullscreen_button=True,
                        height=gallery_height,
                        object_fit="fill",
                    )

    with gr.Tab("Table"):

        @gr.render(inputs=[project_name, dataset])
        def show_experiment_results_table(project_name, dataset):
            if dataset != "":
                gr.DataFrame(
                    experiment_handler.load_experiment_results_for_dataset(
                        project=project_name,
                        dataset=dataset,
                    ),
                    label="Experiment Results",
                    wrap=True,
                    show_search="filter",
                )



import json
from pixaris.feedback_handlers.base import FeedbackHandler
import gradio as gr
from datetime import datetime
import os
import pandas as pd


class LocalFeedbackHandler(FeedbackHandler):
    def __init__(
        self,
        project_feedback_dir: str = "feedback_iterations",
        project_feedback_file: str = "feedback_tracking.jsonl",
        local_feedback_directory: str = "local_results",
    ):
        os.makedirs(local_feedback_directory, exist_ok=True)
        self.local_feedback_directory = local_feedback_directory
        self.project_feedback_dir = project_feedback_dir
        self.project_feedback_file = project_feedback_file
        self.feedback_df = None
        self.feedback_iteration_choices = None
        self.projects = None

    def write_single_feedback(self, feedback: dict) -> None:
        """
        Writes feedback for one image to local feedback storage.

        :param feedback: dict with feedback information. Dict is expected to have the following keys:
            - project: name of the project
            - feedback_iteration: name of the iteration
            - dataset: name of the evaluation set (optional)
            - image_name: name of the image
            - experiment_name: name of the experiment (optional)
            - feedback_indicator: string with feedback value (either "Like", "Dislike", or "Neither")
            - comment: string with feedback comment (optional)
        :type feedback: dict
        """
        # assert non-nullable values are present
        assert all(
            key in feedback.keys()
            for key in [
                "project",
                "feedback_iteration",
                "image_name",
                "feedback_indicator",
            ]
        ), (
            "Missing required feedback keys. Must have 'project', 'feedback_iteration', 'image_name', 'feedback_indicator'"
        )

        # setup row to insert to table
        row_to_insert = {
            "project": feedback["project"],
            "feedback_iteration": feedback["feedback_iteration"],
            "dataset": feedback.get("dataset", ""),
            "image_name": feedback["image_name"],
            "experiment_name": feedback.get("experiment_name", ""),
            "date": datetime.now().isoformat(),
            "comment": feedback.get("comment", ""),
        }

        # determine what to write to feedback columns
        feedback_indicator = feedback["feedback_indicator"]

        if feedback_indicator == "Like":
            row_to_insert["likes"] = 1
            row_to_insert["dislikes"] = 0
        elif feedback_indicator == "Dislike":
            row_to_insert["likes"] = 0
            row_to_insert["dislikes"] = 1
        elif (
            feedback_indicator == "Neither"
        ):  # Is used when uploading images (no feedback given)
            row_to_insert["likes"] = 0
            row_to_insert["dislikes"] = 0
        else:
            raise ValueError(
                "Invalid feedback indicator. Must be 'Like', 'Dislike', or 'Neither'"
            )

        # Construct the directory and file path
        project_dir = os.path.join(
            self.local_feedback_directory,
            feedback["project"],
        )
        os.makedirs(project_dir, exist_ok=True)
        feedback_file_path = os.path.join(project_dir, self.project_feedback_file)

        # Write feedback to the file
        with open(feedback_file_path, "a") as feedback_file:
            feedback_file.write(json.dumps(row_to_insert) + "\n")

        # Display success message
        gr.Info("Feedback saved locally", duration=1)

    def load_projects_list(self) -> list[str]:
        """
        Retrieves list of projects from local storage.

        Returns:
            List of project names.
        """
        projects = [
            d
            for d in os.listdir(self.local_feedback_directory)
            if os.path.isdir(os.path.join(self.local_feedback_directory, d))
        ]
        projects.sort()
        self.projects = projects

        print(f"Found projects: {projects}")
        return projects

    def load_all_feedback_iterations_for_project(self, project: str) -> None:
        """
        Retrieves feedback data for a project from local storage. Adds paths for location of images in
        local directory to the dataframe. Saves the resulting df to self.feedback_df.
        Saves the list of feedback iterations to self.feedback_iteration_choices.

        Args:
            project: str Name of the project

        Returns:
            None
        """
        print(f"Searching locally for feedback data for project {project}...")
        feedback_file_path = os.path.join(
            self.local_feedback_directory,
            project,
            self.project_feedback_file,
        )

        if not os.path.exists(feedback_file_path):
            raise FileNotFoundError(f"No feedback file found at {feedback_file_path}")

        # Load the feedback data from the local JSONL file
        feedback_data = []
        with open(feedback_file_path, "r") as feedback_file:
            for line in feedback_file:
                feedback_data.append(
                    json.loads(line.strip())
                )  # Parse JSONL line into a dictionary

        # Convert the feedback data to a DataFrame
        df = pd.DataFrame(feedback_data)

        # add paths for images to df
        df["image_path_local"] = df.apply(
            lambda row: os.path.join(
                self.local_feedback_directory,
                row["project"],
                self.project_feedback_dir,
                row["feedback_iteration"],
                row["image_name"],
            ),
            axis=1,
        )

        # determine feedback iterations to choose from in this project
        choices = df["feedback_iteration"].unique().tolist()
        choices.sort()

        self.feedback_iteration_choices = choices
        self.feedback_df = df

        print(f"Done. Found feedback iterations: {choices}")

    def load_images_for_feedback_iteration(
        self,
        feedback_iteration: str,
    ) -> list[str]:
        """
        Returns list of local image paths that belong to the feedback iteration.

        Args:
            feedback_iteration: str Name of the feedback iteration

        Returns:
            List of local image paths.
        """
        print(f"Downloading images for feedback iteration {feedback_iteration}...")

        # get relevant data for this feedback iteration
        iteration_df = self.feedback_df.loc[
            self.feedback_df["feedback_iteration"] == feedback_iteration
        ].copy()

        return iteration_df["image_path_local"].tolist()



from typing import List
from pixaris.generation.base import ImageGenerator
from PIL import Image
import os
from io import BytesIO
import vertexai
from vertexai.preview.vision_models import Image as GoogleImage
from vertexai.preview.vision_models import ImageGenerationModel


class Imagen2ImageGenerator(ImageGenerator):
    """
    ImagenGenerator is a class that generates images using the Google Gemini API.
    """

    def __init__(self, gcp_project_id: str, gcp_location: str):
        self.gcp_project_id = gcp_project_id
        self.gcp_location = gcp_location

    def validate_inputs_and_parameters(
        self,
        dataset: List[dict[str, List[dict[str, Image.Image]]]] = [],
        parameters: list[dict[str, str, any]] = [],
    ) -> str:
        """
        Validates the provided dataset and parameters for image generation.

        Args:
            dataset (List[dict[str, List[dict[str, Image.Image]]]]): A list of datasets containing image and mask information.
            parameters (list[dict[str, str, any]]): A list of dictionaries containing generation parameters.

        Raises:
            ValueError: If the validation fails for any reason (e.g., missing fields).
        """
        # Validate dataset
        if not dataset:
            raise ValueError("Dataset cannot be empty.")

        for entry in dataset:
            if not isinstance(entry, dict):
                raise ValueError("Each entry in the dataset must be a dictionary.")

        # Validate parameters, if given
        if parameters:
            for param in parameters:
                if not isinstance(param, dict):
                    raise ValueError("Each parameter must be a dictionary.")  #

    def _encode_image_to_bytes(self, pillow_image: Image.Image) -> bytes:
        """
        Encodes a PIL image to bytes.

        Args:
            pillow_image (Image.Image): The PIL image.

        Returns:
            bytes: Byte representation of the image.
        """
        imgByteArr = BytesIO()
        pillow_image.save(imgByteArr, format=pillow_image.format)
        imgByteArr = imgByteArr.getvalue()
        return imgByteArr

    def _run_imagen(
        self, pillow_images: List[dict], generation_params: List[dict]
    ) -> Image.Image:
        """
        Generates images using the Flux API and checks the status until the image is ready.

        Args:
            pillow_images (List[dict]): A list of dictionaries containing pillow images and mask images.
                Example: [{'node_name': 'Load Input Image', 'pillow_image': <PIL.Image>}, {'node_name': 'Load Mask Image', 'pillow_image': <PIL.Image>}]
            generation_params (list[dict]): A list of dictionaries containing generation params.

        Returns:
            PIL.Image.Image: The generated image.

        Raises:
            requests.exceptions.HTTPError: If the HTTP request returned an unsuccessful status code.
        """

        vertexai.init(project=self.gcp_project_id, location=self.gcp_location)

        prompt = "A beautiful image of a dog"  # PLACEHOLDER

        input_image_path = "input-image.png"
        mask_image_path = "mask-image.png"
        output_image_path = "output-image.png"

        input_image = pillow_images[0]["pillow_image"]
        input_image.save(input_image_path)

        mask_image = pillow_images[1]["pillow_image"]
        mask_image.save(mask_image_path)

        model = ImageGenerationModel.from_pretrained("imagegeneration@006")
        base_img = GoogleImage.load_from_file(location=input_image_path)
        mask_img = GoogleImage.load_from_file(location=mask_image_path)

        images = model.edit_image(
            base_image=base_img,
            mask=mask_img,
            prompt=prompt,
            edit_mode="inpainting-insert",
        )

        images[0].save(output_image_path)
        pillow_image = Image.open(output_image_path)

        os.remove("input-image.png")
        os.remove("mask-image.png")
        os.remove("output-image.png")

        return pillow_image

    def generate_single_image(self, args: dict[str, any]) -> tuple[Image.Image, str]:
        """
        Generates a single image based on the provided arguments.

        Args:
            args (dict[str, any]): A dictionary containing the following keys:
                - pillow_images (list[dict]): A list of dictionaries containing pillow images and mask images.
                - generation_params (list[dict]): A list of dictionaries containing generation params.

        Returns:
            tuple[Image.Image, str]: A tuple containing:
                - image (Image.Image): The generated image.
                - image_name (str): The name of the generated image.
        """
        pillow_images = args.get("pillow_images", [])
        generation_params = args.get("generation_params", [])

        image = self._run_imagen(pillow_images, generation_params)

        # Since the names should all be the same, we can just take the first.
        image_name = pillow_images[0]["pillow_image"].filename.split("/")[-1]

        return image, image_name



import os
import shutil
from google.cloud import storage
from google.cloud.storage import transfer_manager
from pixaris.data_loaders.base import DatasetLoader
from typing import List
from PIL import Image


class GCPDatasetLoader(DatasetLoader):
    """
    GCPDatasetLoader is a class for loading datasets from a Google Cloud Storage bucket. Upon initialisation, the dataset is downloaded to a local directory.

    :param gcp_project_id: The Google Cloud Platform project ID.
    :type gcp_project_id: str
    :param gcp_pixaris_bucket_name: The name of the Google Cloud Storage bucket.
    :type gcp_pixaris_bucket_name: str
    :param project: The name of the project containing the evaluation set.
    :type project: str
    :param dataset: The name of the evaluation set to download images for.
    :type dataset: str
    :param eval_dir_local: The local directory where evaluation images will be saved. Defaults to "local_experiment_inputs".
    :type eval_dir_local: str
    :param force_download: Whether to force download the images even if they already exist locally. Defaults to False.
    :type force_download: bool
    """

    def __init__(
        self,
        gcp_project_id: str,
        gcp_pixaris_bucket_name: str,
        project: str,
        dataset: str,
        eval_dir_local: str = "local_experiment_inputs",
        force_download: bool = True,
    ):
        self.gcp_project_id = gcp_project_id
        self.bucket_name = gcp_pixaris_bucket_name
        self.project = project
        self.dataset = dataset
        self.eval_dir_local = eval_dir_local
        os.makedirs(self.eval_dir_local, exist_ok=True)
        self.force_download = force_download
        self.bucket = None
        self.image_dirs = None

    def _download_dataset(self):
        """
        Downloads evaluation images for a given evaluation set.
        """
        storage_client = storage.Client(project=self.gcp_project_id)
        self.bucket = storage_client.get_bucket(self.bucket_name)
        if self.force_download:
            self._verify_bucket_folder_exists()

        # only download if the local directory does not exist or is empty
        if self._decide_if_download_needed():
            self._download_bucket_dir()

        self.image_dirs = [
            name
            for name in os.listdir(
                os.path.join(self.eval_dir_local, self.project, self.dataset)
            )
            if os.path.isdir(
                os.path.join(self.eval_dir_local, self.project, self.dataset, name)
            )
        ]

    def _verify_bucket_folder_exists(self):
        """
        Verifies that the bucket exists and is not empty.

        :raises: ValueError: If no files are found in the specified directory in the bucket.
        """
        # List the blobs in the bucket. If no blobs are found, raise an error.
        blobs = list(
            self.bucket.list_blobs(
                prefix=f"experiment_inputs/{self.project}/{self.dataset}/"
            )
        )
        if not blobs:
            raise ValueError(
                f"No images found in bucket or bucket does not exist. Please double-check gs://{self.bucket_name}/experiment_inputs/{self.project}/{self.dataset}/."
            )

    def _decide_if_download_needed(self):
        """
        Decides if the download is necessary based on the force_download attribute and existence of the local directory.
        """
        # delete the local directory if force_download is True
        if self.force_download:
            if os.path.exists(
                os.path.join(self.eval_dir_local, self.project, self.dataset)
            ):
                shutil.rmtree(
                    os.path.join(self.eval_dir_local, self.project, self.dataset)
                )

        # Create the local directory if it does not exist
        local_dir = os.path.join(self.eval_dir_local, self.project, self.dataset)
        if os.path.exists(local_dir) and len(os.listdir(local_dir)) > 0:
            return False
        else:
            os.makedirs(local_dir, exist_ok=True)
            return True

    def _download_bucket_dir(self):
        """
        Downloads all files from a specified directory in a Google Cloud Storage bucket to a local directory.
        """
        # Download the blobs to the local directory
        blobs = self.bucket.list_blobs(
            prefix=f"experiment_inputs/{self.project}/{self.dataset}/"
        )
        blob_names = [
            blob.name for blob in blobs if not blob.name.endswith("/")
        ]  # exclude directories

        # remove experiment_inputs/ from the blob names
        blob_names = [name.replace("experiment_inputs/", "") for name in blob_names]

        results = transfer_manager.download_many_to_path(
            self.bucket,
            blob_names,
            blob_name_prefix="experiment_inputs/",
            destination_directory=os.path.join(self.eval_dir_local),
            worker_type=transfer_manager.THREAD,
        )
        # The results list is either `None` or an exception for each blob.
        for name, result in zip(blob_names, results):
            if isinstance(result, Exception):
                print("Failed to download {} due to exception: {}".format(name, result))
            else:
                print(
                    "Downloaded {} to {}.".format(
                        name, os.path.join(self.eval_dir_local, *name.split("/"))
                    )
                )

    def load_dataset(
        self,
    ) -> List[dict[str, List[dict[str, Image.Image]]]]:
        """
        Returns all images in the evaluation set as an iterator of dictionaries containing PIL Images.

        :return: list of dicts containing data loaded from the bucket.
            The key will always be "pillow_images"
            The value is a dict mapping node names to PIL Image objects.
            This dict has a key for each directory in the image_dirs list representing a Node Name.
        :rtype: List[dict[str, List[dict[str, Image.Image]]]]:
        """
        self._download_dataset()
        image_names = self._retrieve_and_check_dataset_image_names()

        dataset = []
        for image_name in image_names:
            pillow_images = []
            for image_dir in self.image_dirs:
                image_path = os.path.join(
                    self.eval_dir_local,
                    self.project,
                    self.dataset,
                    image_dir,
                    image_name,
                )
                # Load the image using PIL
                pillow_image = Image.open(image_path)
                pillow_images.append(
                    {
                        "node_name": f"Load {image_dir.capitalize()} Image",
                        "pillow_image": pillow_image,
                    }
                )
            dataset.append({"pillow_images": pillow_images})
        return dataset

    def _retrieve_and_check_dataset_image_names(self):
        """
        Retrieves the names of the images in the evaluation set and checks if they are the same in each image directory.

        :return: The names of the images in the evaluation set.
        :rtype: List[str]
        :raises: ValueError: If the names of the images in each image directory are not the same.
        """
        basis_names = os.listdir(
            os.path.join(
                self.eval_dir_local, self.project, self.dataset, self.image_dirs[0]
            )
        )
        for image_dir in self.image_dirs:
            image_names = os.listdir(
                os.path.join(self.eval_dir_local, self.project, self.dataset, image_dir)
            )
            if basis_names != image_names:
                raise ValueError(
                    "The names of the images in each image directory should be the same. {} does not match {}.".format(
                        self.image_dirs[0], image_dir
                    )
                )
        return basis_names



from abc import abstractmethod


class FeedbackHandler:
    @abstractmethod
    def write_single_feedback(
        self,
        feedback: dict[str, any],
    ) -> None:
        """Writes feedback for one image to the feedback table."""
        pass

    @abstractmethod
    def load_projects_list(self) -> list[str]:
        """Returns a list of projects."""
        pass

    @abstractmethod
    def load_all_feedback_iterations_for_project(
        self,
        project: str,
    ) -> None:
        """
        Loads all feedback data for all feedback iterations for a project.

        Args:
            project: str Name of the project.
        """
        pass

    @abstractmethod
    def load_images_for_feedback_iteration(
        self,
        feedback_iteration: str,
    ) -> list[str]:
        """
        Loads all images for a feedback iteration. Reutrns a list of image paths that should be rendered
        for this feedback iteration.

        Args:
            feedback_iteration: str Name of the feedback iteration.

        Returns:
            list[str] List of image paths.
        """
        pass



from google.cloud import storage


def upload_workflow_file_to_bucket(
    gcp_project_id: str,
    bucket_name: str,
    dataset: str,
    experiment_run_name: str,
    local_file_path: str,
) -> str:
    """
    Upload a file to a Google Cloud Storage bucket.

    Args:
        gcp_project_id (str): The Google Cloud project ID.
        bucket_name (str): The name of the bucket.
        dataset (str): The name of the evaluation set.
        experiment_run_name (str): The name of the experiment.
        local_file_path (str): The path to the file to upload.
    Returns:
        str: The path to the uploaded file in the bucket.
    """

    storage_client = storage.Client()
    bucket = storage_client.bucket(bucket_name)
    file_name = f"{experiment_run_name}_{local_file_path.split('/')[-1]}"
    destination_path = f"workflows/{dataset}/{file_name}"
    blob = bucket.blob(destination_path)
    blob.upload_from_filename(local_file_path)

    clickable_link = f"https://console.cloud.google.com/storage/browser/_details/{bucket_name}/workflows/{dataset}/{file_name}?project={gcp_project_id}"
    return clickable_link



def merge_dicts(dict1: dict, dict2: dict) -> dict:
    """
    Merge two dictionaries recursively by looping through all keys and updating
    the value lists, so that the resulting dictionary contains all keys and values
    from both dictionaries.
    """
    resulting_dict = {}
    for key, value in dict1.items():
        if key in dict2:
            if isinstance(value, list) and isinstance(dict2[key], list):
                resulting_dict[key] = value + dict2[key]
            else:
                resulting_dict[key] = dict2[key]
        else:
            resulting_dict[key] = value
    for key, value in dict2.items():
        if key not in resulting_dict:
            resulting_dict[key] = value
    return resulting_dict



import os

os.environ["PROTOCOL_BUFFERS_PYTHON_IMPLEMENTATION"] = (
    "python"  # fix for aiplatform, it"s currently using pre 3.20 version of protobuf and we using > 5.0
)
import click
from kubernetes import client, config
import uuid
import time

from pixaris.data_loaders.base import DatasetLoader
from pixaris.generation.base import ImageGenerator
from pixaris.experiment_handlers.base import ExperimentHandler
from pixaris.metrics.base import BaseMetric
from pixaris.orchestration.base import generate_images_based_on_dataset
import pickle
import io
import tarfile
import pathlib
from kubernetes.stream import stream


@click.group()
def cli():
    pass


@cli.command(name="cli-kubernetes-generate-images-based-on-eval-set-execute-remotely")
def cli_kubernetes_generate_images_based_on_dataset_execute_remotely():
    """
    CLI command to trigger remote evaluation. This command should be run on the Kubernetes cluster.
    It will wait for a pickled input file to be uploaded to /tmp/input.pkl and then execute the evaluation.
    """
    pickled_input_path = "/tmp/input.pkl"
    # wait on existence of file (until upload is completed)
    wait_cycles = 0
    while not os.path.exists(f"{pickled_input_path}.done"):
        wait_cycles += 1
        if wait_cycles > 100:
            raise ValueError("Timeout: Input file was not uploaded.")
        print("Waiting for inputs to be uploaded...")
        time.sleep(3)
    with open(pickled_input_path, "rb") as f:
        inputs = pickle.load(f)

    data_loader = inputs["data_loader"]
    image_generator = inputs["image_generator"]
    experiment_handler = inputs["experiment_handler"]
    metrics = inputs["metrics"]
    args = inputs["args"]

    print("Starting generation...")
    generate_images_based_on_dataset(
        data_loader=data_loader,
        image_generator=image_generator,
        experiment_handler=experiment_handler,
        metrics=metrics,
        args=args,
    )


def copy_bytes_to_pod(
    kube_conn, namespace: str, pod_name: str, bytes_to_copy: bytes, dest_path: str
):
    """
    Copy a file to a pod comparable with kubectl cp. See here: https://github.com/kubernetes-client/python/issues/476#issuecomment-2701387338

    Args:
        kube_conn: The connection to the Kubernetes cluster.
        namespace: The namespace of the pod.
        pod_name: The name of the pod.
        bytes_to_copy: The bytes to be copied.
        dest_path: The destination path on the pod.
    """
    # open and compress local tar file
    buf = io.BytesIO()
    with tarfile.open(fileobj=buf, mode="w:gz") as tar:
        tarinfo = tarfile.TarInfo(pathlib.Path(dest_path).name)
        tarinfo.size = len(bytes_to_copy)
        tar.addfile(tarinfo, io.BytesIO(bytes_to_copy))
    buf.seek(0)

    # Copying file
    exec_command = ["tar", "xzvf", "-", "-C", str(pathlib.Path(dest_path).parent)]
    resp = stream(
        kube_conn.connect_get_namespaced_pod_exec,
        pod_name,
        namespace,
        command=exec_command,
        stderr=True,
        stdin=True,
        stdout=True,
        tty=False,
        _preload_content=False,
    )

    chunk_size = 10 * 1024 * 1024
    while resp.is_open():
        resp.update(timeout=1)
        if resp.peek_stdout():
            print(f"STDOUT: {resp.read_stdout()}")
        if resp.peek_stderr():
            print(f"STDERR: {resp.read_stderr()}")
        if read := buf.read(chunk_size):
            resp.write_stdin(read)
        else:
            break
    resp.close()

    resp = stream(
        kube_conn.connect_get_namespaced_pod_exec,
        pod_name,
        namespace,
        command=["touch", f"{dest_path}.done"],
        stderr=True,
        stdin=False,
        stdout=True,
        tty=False,
        _preload_content=False,
    )
    resp.close()

    print(f"File copied to {pod_name}:{dest_path}")


def pixaris_orchestration_kubernetes_locally(
    data_loader: DatasetLoader,
    image_generator: ImageGenerator,
    experiment_handler: ExperimentHandler,
    metrics: list[BaseMetric],
    args: dict[str, any],
    auto_scale: bool,
):
    """
    Trigger remote evaluation on the Kubernetes cluster. This function will pickle the inputs and upload them to the
    Kubernetes cluster. It will then trigger the remote evaluation.

    Args:
    data_loader (DatasetLoader): The data loader to load the evaluation set.
    image_generator (ImageGenerator): The image generator to generate images. E.g. ComfyClusterGenerator
    experiment_handler (ExperimentHandler): The experiment handler to save generated images.
    metrics (list[BaseMetric]): The metrics to calculate.
    args (dict[str, any]): A dictionary of arguments, including:
        - "workflow_apiformat_json" (str): The path to the workflow file in API format.
        - "workflow_pillow_image" (PIL.Image): The image to use as input for the workflow.
        - "dataset" (str): The evaluation set to use.
        - "pillow_images" (list[dict]): A list of images to use as input for the workflow.
        - "experiment_run_name" (str): The name of the run.
        - "max_parallel_jobs" (int): The maximum number of parallel jobs to run.
    auto_scale (bool): Whether to auto scale the cluster to the maximum number of parallel jobs.

    """
    config.load_kube_config()
    if auto_scale:  # always necessary unless the cluster was already scaled up manually
        max_parallel_jobs = args.get("max_parallel_jobs", 1)
        print("Auto scaling...")
        print(f"Scaling to {max_parallel_jobs} replicas")
        apps_v1 = client.AppsV1Api()
        apps_v1.patch_namespaced_deployment_scale(
            "comfy-ui-deployment", "batch", {"spec": {"replicas": max_parallel_jobs}}
        )
    print("Triggering remote evaluation...")

    # Prepare the inputs and and pickle them
    inputs = {
        "data_loader": data_loader,
        "image_generator": image_generator,
        "experiment_handler": experiment_handler,
        "metrics": metrics,
        "args": args,
    }
    pickled_inputs = pickle.dumps(inputs)

    # Trigger the remote evaluation
    batch_v1 = client.BatchV1Api()
    experiment_run_name = args["experiment_run_name"]
    job_name = f"evaluation-{experiment_run_name[0 : (min(30, len(experiment_run_name)))]}-{uuid.uuid4().hex[0:6]}"
    job = {
        "apiVersion": "batch/v1",
        "kind": "Job",
        "metadata": {"name": job_name},
        "spec": {
            "parallelism": 1,
            "completions": 1,
            "backoffLimit": 0,
            "template": {
                "spec": {
                    "serviceAccountName": "experiment",
                    "containers": [
                        {
                            "name": "main",
                            "image": f"europe-west4-docker.pkg.dev/{args['gcp_project_id']}/docker-registry/pixaris_image:latest",
                            "command": [
                                "pixaris-orchestration-kubernetes",
                                "cli-kubernetes-generate-images-based-on-eval-set-execute-remotely",
                            ],
                        }
                    ],
                    "restartPolicy": "Never",
                    "imagePullPolicy": "Always",
                }
            },
        },
    }
    batch_v1.create_namespaced_job(body=job, namespace="batch")

    # Get pod name to upload inputs there
    core_v1 = client.CoreV1Api()
    pod_name = None
    while pod_name is None:
        print("Waiting for pod to start...")
        time.sleep(5)
        job = batch_v1.read_namespaced_job(job_name, "batch")
        if job.status.failed:
            raise ValueError(f"Job {job_name} failed.")

        pods = core_v1.list_namespaced_pod(
            namespace="batch", label_selector=f"job-name={job_name}"
        )
        for pod in pods.items:
            pod_phase = pod.status.phase
            if pod_phase == "Running":
                pod_name = pod.metadata.name

    # Copy the pickled inputs to the pod
    copy_bytes_to_pod(
        kube_conn=core_v1,
        namespace="batch",
        pod_name=pod_name,
        bytes_to_copy=pickled_inputs,
        dest_path="/tmp/input.pkl",
    )

    print("Remote evaluation triggered.")
    print(f"Job logs: kubectl logs -n batch job/{job_name}")
    print(
        f"or https://console.cloud.google.com/kubernetes/job/europe-west4-a/cluster/batch/{job_name}/logs?project={args['gcp_project_id']}"
    )



from typing import Iterable
from PIL import Image


def dict_mean(input_dict_list: Iterable[dict]) -> dict:
    """
    Calculate the mean value for each key in a list of dictionaries.
    Args:
        input_dict_list Iterable[dict]: A list or iterator of dictionaries.
    Returns:
        dict: A dictionary containing the mean value for each key in the input dictionaries.
    """
    # check if all dicts have the same keys
    if len(set(map(frozenset, input_dict_list))) > 1:
        raise ValueError("All dictionaries must have the same keys.")

    means_dict = {}
    for key in input_dict_list[0].keys():
        values = [d[key] for d in input_dict_list]
        means_dict[key] = sum(values) / len(values)
    return means_dict


def normalize_image(image: Image.Image, max_size=(1024, 1024)) -> Image.Image:
    """
    Normalize the given image by placing it on a white background, scaling it while preserving aspect ratio,
    and returning the resulting image.
    Parameters:
    - image (PIL.Image): The input image to be normalized.
    Returns:
    - PIL.Image: The normalized image with a white background.
    """
    # place on white background
    if image.mode != "RGBA":
        image = image.convert("RGBA")

    # Scale image while preserving aspect ratio
    image.thumbnail(max_size)

    # Create white background
    white_base = Image.new("RGB", max_size, (255, 255, 255))

    # Paste object image onto white background
    offset = ((max_size[0] - image.size[0]) // 2, (max_size[1] - image.size[1]) // 2)
    white_base.paste(image, offset, image)
    return white_base



from typing import List
from pixaris.generation.base import ImageGenerator
from pixaris.generation.comfyui import ComfyGenerator
from PIL import Image
import os


from kubernetes import client, config
from threading import Lock, Thread
import requests
import time

DEV_MODE = os.getenv("DEV_MODE", "false") == "true"
mutex = Lock()


class ComfyClusterGenerator(ImageGenerator):
    """
    Cluster to run Comfy workflows. It will automatically fetch available hosts, initiate a new ComfyGenerator for each and distribute the workflows to them.
    If the environment variable DEV_MODE is set to true, it will run the workflows locally and it uses localhost:8188.
    """

    def __init__(
        self,
        workflow_apiformat_json: str,
    ):
        self.workflow_apiformat_json = workflow_apiformat_json
        self.hosts = {}
        self.run_background_task = False

    def _fetch_pod_ips(self) -> list[str]:
        """
        Fetch the IPs of the pods running the Comfy UI in the cluster. If the environment variable DEV_MODE is set to true, it will return localhost.

        Returns:
            list[str]: List of IPs of the pods
        """
        if DEV_MODE:
            return ["127.0.0.1"]
        v1 = client.CoreV1Api()
        pod_list = v1.list_namespaced_pod(
            namespace="batch", label_selector="app=comfy-ui", watch=False
        )
        ips = []
        for pod in pod_list.items:
            ips.append(pod.status.pod_ip)
        print(f"Found Pod IPs: {ips}")
        return ips

    def _fetch_available_hosts(self) -> list[str]:
        """
        Fetch the available hosts by checking if the Comfy UI is running on the pod.

        Returns:
            list[str]: List of available hosts
        """
        hosts = []
        for ip in self._fetch_pod_ips():
            host = f"{ip}:8188"
            try:
                resp = requests.get(f"http://{host}", timeout=5)
                if resp.status_code == 200:
                    hosts.append(host)
            except Exception:
                pass
        return hosts

    def update_available_hosts(self):
        """
        Update the available hosts by fetching the IPs of the pods and checking if the Comfy UI is running on them. Use mutex to avoid conflicts.
        """
        available_hosts = self._fetch_available_hosts()
        global mutex
        with mutex:
            for host in available_hosts:
                if host not in self.hosts:
                    self.hosts[host] = {"in_use": False, "unresponsive": False}
                else:
                    self.hosts[host]["unresponsive"] = False
        print(f"Available hosts: {self.hosts}")

    def _get_host(self) -> str:
        """
        Get an available host. If no host is available, it will wait with a exponential backoff and try again, after the 15th retry, it will raise an exception. Use mutex to avoid conflicts.

        Returns:
            str: The host to use
        """
        global mutex
        for i in range(1, 15 + 1):
            with mutex:
                for host, info in self.hosts.items():
                    if not info["in_use"] and not info["unresponsive"]:
                        self.hosts[host]["in_use"] = True
                        return host
            time.sleep(i * i)
        raise Exception("Timeout for getting host")

    def _release_host(self, host: str):
        """
        Release a host. Use mutex to avoid conflicts.

        Args:
            host (str): The host to release
        """
        global mutex
        with mutex:
            self.hosts[host]["in_use"] = False

    def _mark_host_as_unresponsive(self, host: str):
        """
        Mark a host as unresponsive. Use mutex to avoid conflicts.

        Args:
            host (str): The host to mark as unresponsive
        """
        global mutex
        with mutex:
            self.hosts[host]["unresponsive"] = True

    def start_background_task(self):
        """
        Start a background task to update the available hosts every minute.
        """

        def task():
            while self.run_background_task:
                self.update_available_hosts()
                time.sleep(60)

        self.background_task = Thread(target=task, daemon=True)
        self.background_task.start()

    def close(self):
        """
        Close the cluster and release all hosts.
        """
        self.run_background_task = False

    def validate_inputs_and_parameters(
        self,
        dataset: List[dict[str, List[dict[str, Image.Image]]]] = [],
        parameters: list[dict[str, str, any]] = [],
    ) -> str:
        """
        Creates a dummy generator and then calls ComfyGenerator's validate_inputs_and_parameters method.
        Args:
            dataset (List[dict[str, List[dict[str, Image.Image]]]): A list of dictionaries containing the images to be loaded.
            parameters (list[dict[str, str, any]]): A list of dictionaries containing the parameters to be used for the image generation process.
        """
        dummy_generator = ComfyGenerator(self.workflow_apiformat_json)
        dummy_generator.validate_inputs_and_parameters(dataset, parameters)

    def generate_single_image(self, args: dict[str, any]) -> tuple[Image.Image, str]:
        """
        Generates a single image based on the provided arguments. For this it searches for a host, initialises a ComfyGenerator,
        and lets it modify and execute the workflow to generate the image.
        Args:
            args (dict[str, any]): A dictionary containing the following keys:
            - "workflow_apiformat_json" (str): The path to the workflow file in API format. (ABSOLUTE PATH)!
                    "example.json"
            - "pillow_images" (list[dict]): A dict of [str, Image.Image].
                    The keys should be Node names
                    The values should be the PIL Image objects to be loaded.
                "pillow_images": [{
                    "node_name": "Load Input Image",
                    "pillow_image": Image.new("RGB", (100, 100), color="red"),
                }]
            - "generation_params" (list[dict]): A dictionary of generation_params for the image generation process.
                It should look like this:
                [{
                    "node_name": "GroundingDinoSAMSegment (segment anything)",
                    "input": "prompt",
                    "value": "model, bag, hair",
                }],
        Returns:
            Image.Image: The generated image.
        """
        # on first execution, load config and start background task while ensuring only one worker will do it.
        global mutex
        with mutex:
            if not self.run_background_task:
                if DEV_MODE:
                    config.load_kube_config()
                else:
                    config.load_incluster_config()
                self.run_background_task = True
                self.start_background_task()

        for retry in range(3):
            host = self._get_host()

            # initialize comfy generator for this host
            comfy_generator = ComfyGenerator(
                workflow_apiformat_json=self.workflow_apiformat_json,
                api_host=host,
            )
            try:
                # generate image and release host
                generated_image = comfy_generator.generate_single_image(args)
                self._release_host(host)
                return generated_image

            except Exception as e:
                print(f"Error in ComfyGenerator: {e}")
                self._mark_host_as_unresponsive(host)
                time.sleep((retry + 1) ** 2)
                if retry == 2:
                    raise e

            finally:
                self._release_host(host)



from abc import abstractmethod
from PIL import Image


class ImageGenerator:
    @abstractmethod
    def generate_single_image(self, args: dict[str, any]) -> tuple[Image.Image, str]:
        pass

    def validate_inputs_and_parameters(
        self, inputs: list[dict] = [], parameters: list[dict] = []
    ) -> bool:
        return True



from PIL import Image


def normalize_image(image: Image, max_size=(1024, 1024)) -> Image:
    """
        Normalize the given image by placing it on a white background, scaling it while preserving aspect ratio,
        and returning the resulting image.
        Parameters:
    - image (PIL.Image): The input image to be normalized.
        Returns:
        - PIL.Image: The normalized image with a white background.
    """
    # place on white background
    if image.mode != "RGBA":
        image = image.convert("RGBA")

    # Scale image while preserving aspect ratio
    image.thumbnail(max_size, resample=Image.LANCZOS)

    # Create white background
    white_base = Image.new("RGB", max_size, (255, 255, 255))

    # Paste object image onto white background
    offset = ((max_size[0] - image.size[0]) // 2, (max_size[1] - image.size[1]) // 2)
    white_base.paste(image, offset, image)
    return white_base



from typing import Iterable
from PIL import Image
import pickle
import numpy as np
from google.cloud import storage
from pixaris.experiment_handlers.gcp_tensorboard import GCPTensorboardHandler
from pixaris.experiment_handlers.base import ExperimentHandler


class GCPBucketExperimentHandler(ExperimentHandler):
    def __init__(
        self,
        gcp_project_id: str,
        location: str,
        bucket_name: str,
        bucket_results_path: str,
    ):
        """
        Initializes the GCPBucketHandler.
        Args:
            gcp_project_id (str): The Google Cloud project ID.
            location (str): The Google Cloud location.
            bucket_name (str): The name of the Google Cloud Storage bucket to store the workflow image. If None,
                the workflow image will not be stored in a bucket. Defaults to None.
        """
        self.gcp_project_id = gcp_project_id
        self.location = location
        self.bucket_name = bucket_name
        self.bucket_results_path = bucket_results_path

    def _validate_args(self, args: dict[str, any]):
        # check if all keys are strings
        assert all(isinstance(key, str) for key in args.keys()), (
            "All keys must be strings."
        )

        # check if "pillow_images" is a list of dictionaries containing the correct keys
        if "pillow_images" in args:
            pillow_images = args["pillow_images"]
            assert isinstance(pillow_images, list), "pillow_images must be a list."
            assert all(isinstance(item, dict) for item in pillow_images), (
                "Each item in the list must be a dictionary."
            )
            assert all(
                all(key in item for key in ["node_name", "pillow_image"])
                for item in pillow_images
            ), "Each dictionary must contain the keys 'node_name' and 'pillow_image'."

    def _save_results(
        self,
        experiment_run_name: str,
        results: pickle,
    ):
        """Saves pickled results to a bucket."""

        storage_client = storage.Client()
        bucket = storage_client.bucket(self.bucket_name)

        destination_path = f"{self.bucket_results_path}/{experiment_run_name}.pkl"
        blob = bucket.blob(destination_path)
        blob.upload_from_string(results)
        print(
            f"Results are saved to GCP bucket: 'gs://{self.bucket_name}/{destination_path}"
        )
        print(f"path_in_bucket = '{destination_path}'")

    def store_results(
        self,
        project: str,
        dataset: str,
        experiment_run_name: str,
        image_name_pairs: Iterable[tuple[Image.Image, str]],
        metric_values: dict[str, float],
        args: dict[str, any] = {},
    ):
        """
        Stores the pickled results of an evaluation run to Bucket.

        :param project: The name of the project Will be ignored here.
        :type project: str
        :param dataset: The name of the evaluation set.
        :type dataset: str
        :param experiment_run_name: The name of the run.
        :type experiment_run_name: str
        :param images: A collection of images to log.
        :type images: Iterable[Image.Image]
        :param metrics: A dictionary of metric names and their corresponding values.
        :type metrics: dict[str, float]
        :param args: args given to the ImageGenerator that generated the images.
        :type args: dict[str, any]

        :raises: AssertionError: If any value in the metrics dictionary is not a number.

        """
        self._validate_args(args)
        experiment_run_name = experiment_run_name + str(np.random.randint(99))
        results = {
            "dataset": dataset,
            "experiment_run_name": experiment_run_name,
            "image_name_pairs": image_name_pairs,
            "metric_values": metric_values,
            "args": args,
        }
        pickled_results = pickle.dumps(results)

        self._save_results(experiment_run_name, pickled_results)

    def _write_to_gcp_tensorboard(self, results):
        tensorboard_handler = GCPTensorboardHandler(
            gcp_project_id=self.gcp_project_id,
            location=self.location,
            bucket_name=self.bucket_name,
        )

        tensorboard_handler.store_results(
            dataset=results["dataset"],
            experiment_run_name=results["experiment_run_name"],
            image_name_pairs=results["image_name_pairs"],
            metric_values=results["metric_values"],
            args=results["args"],
        )

    def upload_experiment_from_bucket_to_tensorboard(self):
        """
        Uploads experiment results from a Google Cloud Storage bucket to TensorBoard.
        """

        storage_client = storage.Client()
        bucket = storage_client.bucket(self.bucket_name)
        for blob in bucket.list_blobs(prefix=self.bucket_results_path):
            blob = bucket.blob(blob.name)
            pickled_results = blob.download_as_string()
            results = pickle.loads(pickled_results)

            self._write_to_gcp_tensorboard(results)

            blob.delete()  # cleanup bucket
